# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, unicode_literals, division, absolute_import

import datetime as dt
import decimal
import json
import six

import dateutil.parser

_type_hydrators = {}


def list_model_types():
    return _type_hydrators.keys()


def hydrate_objects(objs, requester=None):
    return [hydrate_object(obj, requester) for obj in objs]


def hydrate_object(obj, requester=None):
    hydrator = _type_hydrators.get(obj["type"], None)
    if not callable(hydrator):
        raise Exception("Can't hydrate a %s!" % obj["type"])
    return hydrator(requester, obj)


class HackerOneEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that leverages an object's `__json__()` method,
    if available, to obtain its default JSON representation.
    """
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return obj.__json__()
        return json.JSONEncoder.default(self, obj)


class _HydrationRegistrant(type):
    def __init__(cls, name, bases, nmspc):
        super(_HydrationRegistrant, cls).__init__(name, bases, nmspc)
        h1_type = getattr(cls, "TYPE", None)
        if h1_type:
            _type_hydrators[h1_type] = cls


@six.add_metaclass(_HydrationRegistrant)
class HackerOneObject(object):
    # subclasses will automatically register themselves
    # as the hydrator for their `TYPE`

    TYPE = None

    def __init__(self, requester, raw_data):
        self._requester = requester
        self._raw_data = raw_data

        self._handle_data()

    @property
    def id(self):
        return self._id

    @property
    def raw_data(self):
        return self._raw_data

    def _handle_data(self):
        assert(self.TYPE == self._raw_data["type"])
        self._id = self._raw_data.get("id", None)
        self._hydrate()

    def _hydrate(self):
        raise NotImplementedError()

    def _hydrate_verbatim(self, val):
        return val

    def _hydrate_datetime(self, val):
        if val is None:
            return None
        return dateutil.parser.parse(val)

    def _hydrate_decimal(self, val):
        return decimal.Decimal(val)

    def _hydrate_list_of_objects(self, val):
        return [self._hydrate_object(x) for x in val]

    def _hydrate_object(self, val):
        return hydrate_object(val, self._requester)

    def _make_attribute(self, attr_name, hydrator, optional=False):
        attrs = self._raw_data["attributes"]
        if optional and attr_name not in attrs:
            attr_val = None
        else:
            attr_val = attrs[attr_name]
        setattr(self, attr_name, hydrator(attr_val))

    def _make_relationship(self, rel_name, hydrator):
        # Some relationships are omitted if non-existent (like assignee)
        rels = self._raw_data["relationships"]
        rel_val = rels.get(rel_name, None)
        if rel_val is not None:
            rel_val = hydrator(rel_val["data"])
        setattr(self, rel_name, rel_val)

    def __eq__(self, other):
        if not isinstance(other, HackerOneObject):
            return False
        if other.TYPE != self.TYPE:
            return False
        if self.id is None:
            return False
        if self.id != other.id:
            return False
        return True

    def __hash__(self):
        if self.id is None:
            return object.__hash__(self)
        return hash((self.TYPE, self.id))

    def __repr__(self):
        if self.id is None:
            return object.__repr__(self)
        return "<%s - %s>" % (self.__class__.__name__, self.id)

    def __json__(self):
        return self.raw_data


class HackerOneResource(HackerOneObject):
    RESOURCE_NAME = None

    def __init__(self, requester, raw_data):
        super(HackerOneResource, self).__init__(requester, raw_data)

    @classmethod
    def make_url(cls, obj_id):
        assert cls.RESOURCE_NAME
        return "/%s/%s" % (cls.RESOURCE_NAME, obj_id)

    @classmethod
    def get(cls, requester, obj_id):
        return requester.request_object(cls.make_url(obj_id))

    @classmethod
    def find(cls, requester, sort=None, yield_pages=False, **kwargs):
        query = {"filter": {}}
        if sort:
            query["sort"] = sort
        for key, val in kwargs.items():
            if isinstance(val, dt.datetime):
                val = val.strftime("%Y-%m-%dT%H:%M:%S.000Z")
            query["filter"][key] = val

        return requester.request_paginated_objects(
            "/%s" % cls.RESOURCE_NAME,
            query,
            yield_pages
        )

    @property
    def url(self):
        return self.make_url(self.id)

    @property
    def _canonical(self):
        return self._raw_data.get("_canonical", False)

    def try_complete(self):
        # TODO: property magic to do this automagically
        # when a field is missing
        if not self._canonical:
            self._fetch_canonical()

    def _fetch_canonical(self):
        self._raw_data = self._requester.request_json(self.url)["data"]
        self._raw_data["_canonical"] = True
        self._hydrate()

    def _hydrate(self):
        raise NotImplementedError()


class Report(HackerOneResource):
    TYPE = "report"
    RESOURCE_NAME = "reports"

    STATES = {
        "new",
        "triaged",
        "needs-more-info",
        "resolved",
        "not-applicable",
        "informative",
        "duplicate",
        "spam",
    }

    @property
    def total_bounty(self):
        return sum(b.amount for b in self.bounties)

    @property
    def total_payout(self):
        return sum(b.amount + b.bonus_amount for b in self.bounties)

    @property
    def time_to_closed(self):
        if not self.closed_at:
            return None
        return self.closed_at - self.created_at

    @property
    def time_to_first_response(self):
        if not self.first_program_activity_at:
            return None
        return self.first_program_activity_at - self.created_at

    @property
    def html_url(self):
        return "https://hackerone.com/reports/%s" % self.id

    def _hydrate(self):
        self._make_attribute("created_at", self._hydrate_datetime)
        self._make_attribute("last_activity_at", self._hydrate_datetime)
        self._make_attribute("last_program_activity_at", self._hydrate_datetime)
        self._make_attribute("last_reporter_activity_at", self._hydrate_datetime)
        self._make_attribute("triaged_at", self._hydrate_datetime)
        self._make_attribute("swag_awarded_at", self._hydrate_datetime)
        self._make_attribute("bounty_awarded_at", self._hydrate_datetime)
        self._make_attribute("closed_at", self._hydrate_datetime)
        self._make_attribute("disclosed_at", self._hydrate_datetime)
        self._make_attribute("title", self._hydrate_verbatim)
        self._make_attribute("first_program_activity_at", self._hydrate_datetime)
        self._make_attribute("vulnerability_information", self._hydrate_verbatim)
        self._make_attribute("state", self._hydrate_verbatim)

        self._make_relationship("reporter", self._hydrate_object)
        self._make_relationship("assignee", self._hydrate_object)
        self._make_relationship("program", self._hydrate_object)
        self._make_relationship("bounties", self._hydrate_list_of_objects)
        self._make_relationship("activities", self._hydrate_list_of_objects)
        self._make_relationship("attachments", self._hydrate_list_of_objects)
        self._make_relationship("swag", self._hydrate_list_of_objects)
        self._make_relationship("vulnerability_types", self._hydrate_list_of_objects)
        self._make_relationship("summaries", self._hydrate_list_of_objects)


class User(HackerOneObject):
    TYPE = "user"

    def _hydrate(self):
        self._make_attribute("username", self._hydrate_verbatim)
        self._make_attribute("name", self._hydrate_verbatim)
        self._make_attribute("disabled", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)
        self._make_attribute("profile_picture", self._hydrate_verbatim)

    def __str__(self):
        return self.username

    def __repr__(self):
        return "<%s - %s>" % (self.__class__.__name__, self.username)


class Program(HackerOneObject):
    TYPE = "program"

    def _hydrate(self):
        self._make_attribute("handle", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)
        self._make_attribute("updated_at", self._hydrate_datetime)


class Bounty(HackerOneObject):
    TYPE = "bounty"

    def _hydrate(self):
        self._make_attribute("created_at", self._hydrate_datetime)
        self._make_attribute("bonus_amount", self._hydrate_decimal)
        self._make_attribute("amount", self._hydrate_decimal)


class VulnerabilityType(HackerOneObject):
    TYPE = "vulnerability-type"

    def _hydrate(self):
        self._make_attribute("name", self._hydrate_verbatim)
        self._make_attribute("description", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)


class Attachment(HackerOneObject):
    TYPE = "attachment"

    def _hydrate(self):
        self._make_attribute("file_name", self._hydrate_verbatim)
        self._make_attribute("content_type", self._hydrate_verbatim)
        self._make_attribute("file_size", self._hydrate_verbatim)
        self._make_attribute("expiring_url", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)


class ReportSummary(HackerOneObject):
    TYPE = "report-summary"

    def _hydrate(self):
        self._make_attribute("content", self._hydrate_verbatim)
        self._make_attribute("category", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)
        self._make_attribute("updated_at", self._hydrate_datetime)

        self._make_relationship("user", self._hydrate_object)


class Group(HackerOneObject):
    TYPE = "group"

    def _hydrate(self):
        self._make_attribute("name", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)


class Swag(HackerOneObject):
    TYPE = "swag"

    def _hydrate(self):
        self._make_attribute("sent", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)

        self._make_relationship("address", self._hydrate_object)


class Address(HackerOneObject):
    TYPE = "address"

    def _hydrate(self):
        self._make_attribute("name", self._hydrate_verbatim)
        self._make_attribute("street", self._hydrate_verbatim)
        self._make_attribute("city", self._hydrate_verbatim)
        self._make_attribute("postal_code", self._hydrate_verbatim)
        self._make_attribute("state", self._hydrate_verbatim)
        self._make_attribute("country", self._hydrate_verbatim)
        self._make_attribute("tshirt_size", self._hydrate_verbatim)
        self._make_attribute("phone_number", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)


class ActivityBase(HackerOneObject):
    def _hydrate(self):
        self._make_attribute("message", self._hydrate_verbatim)
        self._make_attribute("internal", self._hydrate_verbatim)
        self._make_attribute("created_at", self._hydrate_datetime)
        self._make_attribute("updated_at", self._hydrate_datetime)

        self._make_relationship("actor", self._hydrate_object)
        self._make_relationship("attachments", self._hydrate_list_of_objects)

        self._activity_hydrate()

    def _activity_hydrate(self):
        pass


class ActivityStateChange(ActivityBase):
    pass


class ActivityPublicAgreement(ActivityBase):
    TYPE = "activity-agreed-on-going-public"


class ActivityBountyBase(ActivityBase):
    def _activity_hydrate(self):
        self._make_attribute("bounty_amount", self._hydrate_decimal)
        self._make_attribute("bonus_amount", self._hydrate_decimal)


class ActivityBountyAwarded(ActivityBase):
    TYPE = "activity-bounty-awarded"


class ActivityBountySuggested(ActivityBountyBase):
    TYPE = "activity-bounty-suggested"


class ActivityBugCloned(ActivityBase):
    TYPE = "activity-bug-cloned"

    def _activity_hydrate(self):
        # TODO: The API documentation seems to disagree with
        # what's actually being sent back here?
        self._make_attribute("original_report_id", self._hydrate_verbatim, optional=True)


class ActivityBugDuplicate(ActivityStateChange):
    TYPE = "activity-bug-duplicate"

    def _activity_hydrate(self):
        self._make_attribute("original_report_id", self._hydrate_verbatim, optional=True)


class ActivityBugInformative(ActivityStateChange):
    TYPE = "activity-bug-informative"


class ActivityBugNeedsMoreInfo(ActivityStateChange):
    TYPE = "activity-bug-needs-more-info"


class ActivityBugNew(ActivityStateChange):
    TYPE = "activity-bug-new"


class ActivityBugNotApplicable(ActivityStateChange):
    TYPE = "activity-bug-not-applicable"


class ActivityBugReopened(ActivityStateChange):
    TYPE = "activity-bug-reopened"


class ActivityBugResolved(ActivityStateChange):
    TYPE = "activity-bug-resolved"


class ActivityBugSpam(ActivityStateChange):
    TYPE = "activity-bug-spam"


class ActivityBugTriaged(ActivityStateChange):
    TYPE = "activity-bug-triaged"


class ActivityComment(ActivityBase):
    TYPE = "activity-comment"


class ActivityExternalUserInvitationCancelled(ActivityBase):
    TYPE = "activity-external-user-invitation-cancelled"

    def _activity_hydrate(self):
        self._make_attribute("email", self._hydrate_verbatim)


class ActivityExternalUserInvited(ActivityBase):
    TYPE = "activity-external-user-invited"

    def _activity_hydrate(self):
        self._make_attribute("email", self._hydrate_verbatim)


class ActivityExternalUserJoined(ActivityBase):
    TYPE = "activity-external-user-joined"

    def _activity_hydrate(self):
        self._make_attribute("duplicate_report_id", self._hydrate_verbatim)


class ActivityExternalUserRemoved(ActivityBase):
    TYPE = "activity-external-user-removed"

    def _activity_hydrate(self):
        self._make_relationship("removed_user", self._hydrate_object)


class ActivityGroupAssignedToBug(ActivityBase):
    TYPE = "activity-group-assigned-to-bug"

    def _activity_hydrate(self):
        self._make_relationship("group", self._hydrate_object)


class ActivityHackerRequestedMediation(ActivityBase):
    TYPE = "activity-hacker-requested-mediation"


class ActivityManuallyDisclosed(ActivityBase):
    TYPE = "activity-manually-disclosed"


class ActivityMediationRequested(ActivityBase):
    TYPE = "activity-mediation-requested"


class ActivityNotEligibleForBounty(ActivityBase):
    TYPE = "activity-not-eligible-for-bounty"


class ActivityReferenceIDAdded(ActivityBase):
    TYPE = "activity-reference-id-added"

    def _activity_hydrate(self):
        self._make_attribute("reference", self._hydrate_verbatim)
        self._make_attribute("reference_url", self._hydrate_verbatim)


class ActivityReportBecamePublic(ActivityBase):
    TYPE = "activity-report-became-public"


class ActivityReportTitleUpdated(ActivityBase):
    TYPE = "activity-report-title-updated"


class ActivityReportVulnerabilityTypesUpdated(ActivityBase):
    TYPE = "activity-report-vulnerability-types-updated"

    def _activity_hydrate(self):
        self._make_relationship("old_vulnerability_types", self._hydrate_list_of_objects)
        self._make_relationship("new_vulnerability_types", self._hydrate_list_of_objects)


class ActivitySwagAwarded(ActivityBase):
    TYPE = "activity-swag-awarded"

    def _activity_hydrate(self):
        self._make_relationship("swag", self._hydrate_object)


class ActivityUserAssignedToBug(ActivityBase):
    TYPE = "activity-user-assigned-to-bug"

    def _activity_hydrate(self):
        self._make_relationship("assigned_user", self._hydrate_object)


class ActivityUserBannedFromProgram(ActivityBase):
    TYPE = "activity-user-banned-from-program"

    def _activity_hydrate(self):
        self._make_relationship("removed_user", self._hydrate_object)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, unicode_literals, division, absolute_import

from requests.auth import HTTPBasicAuth
from requests_futures.sessions import FuturesSession
from six.moves.urllib.parse import quote

from .models import hydrate_object, hydrate_objects
from .lazy_listing import LazyListing
from . import __version__


def encode_params(obj, keys=()):
    if isinstance(obj, dict):
        # key[bar]=1&key[baz]=2
        return "&".join(encode_params(val, keys + (key,)) for key, val in obj.items())
    # key[foo][]=1&key[foo][]=2
    elif isinstance(obj, (list, tuple)):
        return "&".join(encode_params(val, keys + ("",)) for val in obj)
    # `obj` is just a value, key=1
    else:
        encoded_keys = ""
        for depth, key in enumerate(keys):
            # All keys but top-level keys should be in brackets, i.e.
            # `key[foo]=1`, not `[key][foo]=1`
            encoded_keys += key if depth == 0 else "[" + key + "]"
        return quote(encoded_keys) + "=" + quote(obj)


class HackerOneClient(object):
    BASE_URL = "https://api.hackerone.com/v1"
    REQUEST_HEADERS = {
        "User-Agent": "HackerOne Python Client v" + __version__,
    }

    def __init__(self, identifier, token):
        self.identifier = identifier
        self.token = token
        self._init_session()

    def _init_session(self):
        self.s = FuturesSession()
        self.s.headers.update(self.REQUEST_HEADERS)
        self.s.auth = HTTPBasicAuth(self.identifier, self.token)

    def make_request(self, url, params=None, data=None, method=None):
        if method is None:
            method = "GET"
        if not url.startswith("http"):
            url = self.BASE_URL + url
        if isinstance(params, dict):
            params = encode_params(params)

        return self.s.request(method, url, params=params, data=data)

    def request_json(self, url, params=None, data=None, method=None):
        r = self.make_request(url, params, data, method).result()
        r.raise_for_status()
        return r.json()

    def request_object(self, url, params=None, data=None, method=None):
        data = self.request_json(url, params, data, method)["data"]
        # If we're fetching a single object make sure that consumers
        # know this is the canonical version
        data["_canonical"] = True
        return hydrate_object(data, self)

    def request_paginated_objects(self, url, params=None, yield_pages=False):
        future = self.make_request(url, params)
        res_iter = self._request_paginated_inner(url, future, yield_pages)
        return LazyListing(res_iter)

    def _request_paginated_inner(self, url, future, yield_pages):
        while url:
            resp = future.result()
            resp.raise_for_status()
            parsed = resp.json()
            url = parsed["links"].get('next')
            if url:
                future = self.make_request(url)

            hydrated_objs = hydrate_objects(parsed["data"], self)
            if yield_pages:
                yield hydrated_objs
            else:
                for obj in hydrated_objs:
                    yield obj

    def get_resource(self, rsrc_type, obj_id):
        return rsrc_type.get(self, obj_id)

    def find_resources(self, rsrc_type, sort=None, yield_pages=False, **kwargs):
        """Find instances of `rsrc_type` that match the filter in `**kwargs`"""
        return rsrc_type.find(self, sort=sort, yield_pages=yield_pages, **kwargs)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, unicode_literals, division, absolute_import

import collections
import itertools
import threading

import six


# From itertools documentation
def _consume(iterator, n=None):
    """Advance the iterator n-steps ahead. If n is none, consume entirely."""
    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        collections.deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(itertools.islice(iterator, n, n), None)


def _slice_required_len(slice_obj):
    """
    Calculate how many items must be in the collection to satisfy this slice

    returns `None` for slices may vary based on the length of the underlying collection
    such as `lst[-1]` or `lst[::]`
    """
    if slice_obj.step and slice_obj.step != 1:
        return None
    # (None, None, *) requires the entire list
    if slice_obj.start is None and slice_obj.stop is None:
        return None

    # Negative indexes are hard without knowing the collection length
    if slice_obj.start and slice_obj.start < 0:
        return None
    if slice_obj.stop and slice_obj.stop < 0:
        return None

    if slice_obj.stop:
        if slice_obj.start and slice_obj.start > slice_obj.stop:
            return 0
        return slice_obj.stop
    return slice_obj.start + 1


class LazyListing(object):
    def __init__(self, res_iter):
        self._items = []
        self._res_iter = iter(res_iter)
        self._res_lock = threading.RLock()

    def _ensure_elements_available(self, n=None):
        if self._res_iter:
            with self._res_lock:
                if not self._res_iter:
                    return
                if n is None or n > len(self._items):
                    _consume(self, n)

    def __iter__(self):
        if self._res_iter:
            with self._res_lock:
                # Are there still items left in the resource iterator?
                if self._res_iter:
                    # Yield any items we already have
                    for item in self._items:
                        yield item
                    # Start consuming items from the resource iterator
                    for item in self._res_iter:
                        self._items.append(item)
                        yield item
                    self._res_iter = None
                    return

        for item in self._items:
            yield item

    def __len__(self):
        self._ensure_elements_available()
        return len(self._items)

    def __getitem__(self, item):
        # Figure out how many items we need to have in self._items to satisfy this.
        # This ensures we don't need to consume the entire iterator to get a
        # slice of the first 2 elems, for ex.
        required_len = None
        if isinstance(item, slice):
            required_len = _slice_required_len(item)
        elif isinstance(item, six.integer_types):
            required_len = item + 1
            if required_len <= 0:
                required_len = None
        self._ensure_elements_available(required_len)
        return self._items[item]

    def __contains__(self, item):
        self._ensure_elements_available()
        return item in self._items

    def __json__(self):
        self._ensure_elements_available()
        return self._items



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, unicode_literals, division, absolute_import

from collections import OrderedDict
import itertools
import json
import unittest

import requests_mock

from h1.client import encode_params, HackerOneClient
from h1.models import Report
from . import load_resource_blob


class TestParamEncoding(unittest.TestCase):

    def test_simple_val(self):
        # Make sure we use `OrderedDict` so we don't depend on
        # implementation-defined iteration order
        params = encode_params(OrderedDict((
            ("foo", "bar"),
            ("baz", "quux"),
        )))
        self.assertEqual("foo=bar&baz=quux", params)

    def test_list_val(self):
        params = encode_params(OrderedDict((
            ("foo", "bar"),
            ("baz", ("1", "2")),
        )))
        self.assertEqual("foo=bar&baz%5B%5D=1&baz%5B%5D=2", params)

    def test_dict_val(self):
        params = encode_params(OrderedDict((
            ("foo", "bar"),
            ("baz", OrderedDict((
                ("plugh", "foozy"),
                ("quuxy", "bazzy"),
            ))),
        )))
        self.assertEqual("foo=bar&baz%5Bplugh%5D=foozy&baz%5Bquuxy%5D=bazzy", params)


class TestHackerOneClient(unittest.TestCase):

    def test_request_json(self):
        report_blob = load_resource_blob("responses/report")
        client = HackerOneClient("a", "b")
        json_url = HackerOneClient.BASE_URL + "/reports/1"
        with requests_mock.mock() as m:
            m.get(json_url, text=json.dumps(report_blob))

            report_json = client.request_json("/reports/1")
            self.assertDictEqual(report_blob, report_json)
            # Make sure absolute URLs work as well
            report_json = client.request_json(json_url)
            self.assertDictEqual(report_blob, report_json)

    def test_get_resource(self):
        report_text = json.dumps(load_resource_blob("responses/report"))
        with requests_mock.mock() as m:
            m.get(HackerOneClient.BASE_URL + "/reports/1", text=report_text)

            client = HackerOneClient("a", "b")
            report = client.get_resource(Report, 1)
            self.assertEqual(report.title, "XSS in login form")

    def test_find_resources(self):
        listing_text = json.dumps(load_resource_blob("responses/report-list-onepage"))
        with requests_mock.mock() as m:
            filters = {"foo_gt": "bar"}
            filter_str = "filter%5Bprogram%5D%5B%5D=foo&filter%5Bfoo_gt%5D=bar"
            m.get(HackerOneClient.BASE_URL + "/reports?" + filter_str, text=listing_text)

            client = HackerOneClient("a", "b")
            listing = client.find_resources(Report, program=["foo"], **filters)

            # The listing should contain two reports
            self.assertEqual(len(listing), 2)
            self.assertTrue(all(isinstance(r, Report) for r in listing))

    def test_find_resources_yield_pages(self):
        page_one = json.dumps(load_resource_blob("responses/report-list-twopage-0"))
        page_two = json.dumps(load_resource_blob("responses/report-list-twopage-1"))
        with requests_mock.mock() as m:
            filter_str = "filter%5Bprogram%5D%5B%5D=foo"
            m.get(HackerOneClient.BASE_URL + "/reports?" + filter_str, text=page_one)
            m.get(HackerOneClient.BASE_URL + "/reports/nextpage", text=page_two)

            client = HackerOneClient("a", "b")
            listing = client.find_resources(Report, program=["foo"], yield_pages=True)

            # The listing should contain two pages each with one report
            self.assertEqual(len(listing), 2)
            # Flatten the pages into a single page
            all_items = list(itertools.chain(*listing))
            self.assertEqual(len(all_items), 2)
            self.assertTrue(all(isinstance(r, Report) for r in all_items))



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, unicode_literals, division, absolute_import

import itertools
import unittest

import six.moves

from h1.lazy_listing import LazyListing


class TestLazyListing(unittest.TestCase):

    @staticmethod
    def _gen_lazy_listing():
        return LazyListing(six.moves.range(4))

    def test_iteration(self):
        listing = self._gen_lazy_listing()
        # `list()` implicitly calls `__iter__()`
        self.assertListEqual(list(listing), [0, 1, 2, 3])

    def test_partial_iteration(self):
        listing = self._gen_lazy_listing()
        # stop iterating after 2 item
        for i in itertools.islice(listing, 2):
            pass

        # Now restart iteration and make sure we get the whole list
        self.assertListEqual(list(listing), [0, 1, 2, 3])

    def test_contains(self):
        listing = self._gen_lazy_listing()
        self.assertFalse(5 in listing)
        self.assertTrue(1 in listing)

    def test_simple_index(self):
        listing = self._gen_lazy_listing()
        self.assertEqual(listing[1], 1)

    def test_negative_index(self):
        listing = self._gen_lazy_listing()
        self.assertEqual(listing[-1], 3)

    def test_slice_start(self):
        listing = self._gen_lazy_listing()
        self.assertEqual(listing[1::], [1])

    def test_slice_negative_start(self):
        listing = self._gen_lazy_listing()
        self.assertEqual(listing[-1::], [3])

    def test_slice_stop(self):
        listing = self._gen_lazy_listing()
        self.assertListEqual(listing[:2], [0, 1])

    def test_slice_negative_stop(self):
        listing = self._gen_lazy_listing()
        self.assertListEqual(listing[:-1], [0, 1, 2])

    def test_slice_start_stop(self):
        listing = self._gen_lazy_listing()
        self.assertListEqual(listing[1:2], [1])

    def test_slice_equal_start_stop(self):
        listing = self._gen_lazy_listing()
        self.assertListEqual(listing[1:1:1], [])

    def test_slice_copy(self):
        listing = self._gen_lazy_listing()
        self.assertListEqual(listing[::], [0, 1, 2, 3])

    def test_slice_reverse(self):
        listing = self._gen_lazy_listing()
        self.assertListEqual(listing[::-1], [3, 2, 1, 0])

    def test_len(self):
        listing = self._gen_lazy_listing()
        self.assertEqual(len(listing), 4)

    def test_json_serialization(self):
        listing = self._gen_lazy_listing()
        as_json = listing.__json__()
        self.assertListEqual(as_json, [0, 1, 2, 3])
        self.assertIsInstance(as_json, list)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function

import re
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

with open('h1/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')


def read_long_description(filename="README.rst"):
    with open(filename) as f:
        return f.read().strip()


def requirements(filename="requirements.in"):
    with open(filename) as f:
        return f.readlines()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

setup(
    author='Uber Product Security',
    description='A HackerOne API client for Python',
    include_package_data=True,
    install_requires=requirements(),  # Edit these in `requirements.in`.
    tests_require=requirements('requirements-test.in'),
    cmdclass={'test': PyTest},
    license='MIT',
    long_description=read_long_description(),
    name='h1',
    packages=find_packages(),
    url='https://github.com/uber-common/h1-python',
    version=version,
    zip_safe=False,
    classifiers=[
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ]
)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, unicode_literals, division, absolute_import

import datetime as dt
from decimal import Decimal
import json
import unittest

from h1.models import HackerOneEncoder, hydrate_object, Group, Report, list_model_types
from . import load_resource_blob


class TestModelEncoding(unittest.TestCase):

    def test_serialize_round_trip(self):
        # A model should deserialize to the same JSON as it was
        # originally serialized from
        group_blob = load_resource_blob("group")
        group = hydrate_object(group_blob)
        self.assertIsInstance(group, Group)
        round_trip_group_blob = json.loads(json.dumps(group, cls=HackerOneEncoder))
        self.assertDictEqual(round_trip_group_blob, group_blob)

    def test_serialize_round_trip_list(self):
        group_blob = load_resource_blob("group")
        group = hydrate_object(group_blob)
        group_list = [group, group]
        round_trip_group_list = json.loads(json.dumps(group_list, cls=HackerOneEncoder))
        self.assertEqual(len(round_trip_group_list), 2)
        self.assertDictEqual(round_trip_group_list[0], group_blob)


class TestModelHydration(unittest.TestCase):
    def test_type_mismatch_throws(self):
        with self.assertRaises(AssertionError):
            Report(None, load_resource_blob("group"))

    def test_report_properties(self):
        report = Report(None, load_resource_blob("report"))
        self.assertEqual(report.id, "1337")
        self.assertEqual(report.total_bounty, Decimal("1000.00"))
        self.assertEqual(report.total_payout, Decimal("1100.00"))
        self.assertEqual(report.html_url, "https://hackerone.com/reports/1337")
        self.assertEqual(report.time_to_first_response, dt.timedelta(days=1))
        self.assertEqual(report.time_to_closed, dt.timedelta(days=2))

    def test_hydrate_all_objects(self):
        # Try to hydrate every model we know about from the test resources
        for model_type in list_model_types():
            hydrate_object(load_resource_blob(model_type))


class TestModelSemantics(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        group_1_blob = load_resource_blob("group")
        cls.group_1a = hydrate_object(group_1_blob)
        cls.group_1b = hydrate_object(group_1_blob)
        group_2_blob = load_resource_blob("group")
        group_2_blob["id"] = "1338"
        cls.group_2 = hydrate_object(group_2_blob)
        cls.report = hydrate_object(load_resource_blob("report"))

    def test_same_id_equal(self):
        self.assertEqual(self.group_1a, self.group_1b)

    def test_different_id_not_equal(self):
        self.assertNotEqual(self.group_1a, self.group_2)

    def test_different_type_not_equal(self):
        self.assertNotEqual(self.group_1a, self.report)

    def test_different_none_not_equal(self):
        self.assertNotEqual(self.group_1a, None)
